import sys
from typing import List
from pathlib import Path

from ehelply_bootstrapper.drivers.config import Config
from ehelply_bootstrapper.drivers.fast_api import Fastapi
from ehelply_bootstrapper.drivers.mongo import Mongo
from ehelply_bootstrapper.drivers.redis import Redis
from ehelply_bootstrapper.drivers.sentry import Sentry
from ehelply_bootstrapper.drivers.socketio import Socketio
from ehelply_bootstrapper.drivers.rabbitmq import Rabbitmq
from ehelply_bootstrapper.drivers.aws import AWS
from ehelply_bootstrapper.utils.connection_details import ConnectionDetails
from ehelply_bootstrapper.utils.state import State

from ehelply_logger.Logger import Logger

LOADABLE_FASTAPI = "fastapi"
LOADABLE_MONGO = "mongo"
LOADABLE_REDIS = "redis"
LOADABLE_SENTRY = "sentry"
LOADABLE_SOCKET = "socket"
LOADABLE_RABBITMQ = "rabbitmq"
LOADABLE_AWS = "aws"


class Bootstrap:
    """
    Bootstrap class gets a service ready to be loaded

    This class SHOULD be overridden to ensure maximum control
    """

    def __init__(self, service_name: str, service_key: str, service_version: str, loadables: List[str],
                 config_path: str = None,
                 configs: List[str] = None,
                 dev_mode: bool = False):
        self.service_name: str = service_name
        self.service_key: str = service_key
        self.config_path: str = config_path
        self.dev_mode: bool = dev_mode
        self.loadables: List[str] = loadables
        self.service_version: str = service_version

        # Applications and clients
        self.fastapi_driver: Fastapi = None
        self.socket_driver: Socketio = None
        self.redis_driver: Redis = None
        self.mongo_driver: Mongo = None
        self.rabbitmq_driver: Rabbitmq = None
        self.aws_driver: Rabbitmq = None

        if len(sys.argv) > 1 and "--dev" in sys.argv:
            self.dev_mode = True

        if self.config_path is None:
            if True or self.dev_mode:
                self.config_path = str(Path(__file__).parents[1]) + "/config"
            else:
                self.config_path = "/config/" + self.service_key

        self.logger: Logger = Logger(debug_mode=self.dev_mode)

        self.logger.info("Starting loading procedure for `" + self.service_name + "` with key `" + self.service_key + "`")
        self.logger.info(" * Dev mode: " + str(self.dev_mode))
        self.logger.info(" * Configuration path: " + str(self.config_path))
        self.logger.info(" * Loading configs: " + str(configs))
        self.logger.info(" * Loading drivers: " + str(self.loadables))

        self.check_loadables_conflicts()
        self.logger.info("Confliction test passed")

        self.check_loadables_improvements()

        self.logger.info("Pre loading...")
        self.pre_load()

        self.logger.info("Loading configuration...")
        Config(self.service_key, self.config_path, configs, dev_mode=self.dev_mode).init()

        if LOADABLE_SENTRY in self.loadables:
            self.logger.info("Loading sentry...")
            self.sentry_init()

        if LOADABLE_FASTAPI in self.loadables:
            self.logger.info("Loading fast api...")
            self.fastapi_init()
            self.logger.info("Registering middleware to fast api...")
            self.fastapi_middleware()
            self.logger.info("Registering routers to fast api...")
            self.fastapi_routers()

        if LOADABLE_SOCKET in self.loadables:
            self.logger.info("Loading socket io...")
            self.socket_init()

        if LOADABLE_REDIS in self.loadables:
            self.logger.info("Loading redis...")
            self.redis_init()

        if LOADABLE_MONGO in self.loadables:
            self.logger.info("Loading mongo...")
            self.mongo_init()

        if LOADABLE_RABBITMQ in self.loadables:
            self.logger.info("Loading rabbitmq...")
            self.rabbitmq_init()

        if LOADABLE_AWS in self.loadables:
            self.logger.info("Loading AWS...")
            self.aws_init()

        if LOADABLE_FASTAPI in self.loadables:
            self.logger.info("Registering additional endpoints to fast api...")
            self.fastapi_register_endpoints()

        if LOADABLE_SOCKET in self.loadables:
            self.logger.info("Registering additional events to socket io...")
            self.socket_register_events()

        self.logger.info("Setting up service state...")
        State.logger = self.logger
        State.app = self.fastapi_driver
        State.sockets = self.socket_driver
        State.redis = self.redis_driver
        State.mongo = self.mongo_driver
        State.rabbitmq = self.rabbitmq_driver
        State.aws = self.aws_driver

        self.logger.info("Ready")

        self.logger.info("Running post load...")
        self.post_load()

        if (LOADABLE_FASTAPI in self.loadables) and self.dev_mode:
            self.logger.info("Starting fastapi dev server with uvicorn")
            self.fastapi_driver.run_dev_server()

    def check_loadables_conflicts(self):
        if LOADABLE_SOCKET in self.loadables and LOADABLE_FASTAPI not in self.loadables:
            raise Exception("Cannot use sockets without fastapi.")

    def check_loadables_improvements(self):
        if LOADABLE_SENTRY not in self.loadables:
            self.logger.info("You're not using sentry. Consider using sentry to improve error catching and debugging.")

    def pre_load(self):
        pass

    def post_load(self):
        pass

    def fastapi_init(self):
        """
        Create the fast api app object
        :return:
        """
        self.fastapi_driver = Fastapi(service_name=self.service_name, service_version=self.service_version,
                                      dev_mode=self.dev_mode).init()

    def fastapi_middleware(self):
        """
        Inject middleware into fast api
        :return:
        """
        # Cors
        self.fastapi_driver.cors(origins=['*'], allow_credentials=True)

        # Zipping Responses
        self.fastapi_driver.compression(min_size=500)

        if LOADABLE_SENTRY in self.loadables:
            self.fastapi_driver.sentry()

    def fastapi_routers(self):
        """
        Register routers to fastapi

        Example
        -------
        self.fastapi_app.include_router(
            admin.router,
            prefix="/admin",
            tags=["admin"],
            responses={404: {"description": "Not found"}},
        )

        :return:
        """
        pass

    def fastapi_register_endpoints(self):
        """
        This allows you to import other files with extra fastapi endpoints which were not inside of routers
        :return:
        """
        pass

    def sentry_init(self):
        """
        Register sentry
        :return:
        """
        Sentry(dev_mode=self.dev_mode).init()

    def socket_init(self):
        """
        This sets up the sockets app

        :return:
        """
        self.socket_driver = Socketio(dev_mode=self.dev_mode).init()
        self.fastapi_driver.mount_app("/sockets", self.socket_driver.socket_app)

    def socket_register_events(self):
        """
        This allows you to import other files and/or define events for socket io
        :return:
        """
        pass

    def redis_init(self):
        self.redis_driver = Redis(dev_mode=self.dev_mode).init()

    def mongo_init(self):
        self.mongo_driver = Mongo(dev_mode=self.dev_mode).init()

    def aws_init(self):
        self.aws_driver = AWS(dev_mode=self.dev_mode).init()

    def rabbitmq_init(self):
        self.rabbitmq_driver = Rabbitmq(dev_mode=self.dev_mode,
                                        connection_details=ConnectionDetails(host=State.config.bootstrap.rabbitmq.host,
                                                                             port=State.config.bootstrap.rabbitmq.port,
                                                                             username=State.config.bootstrap.rabbitmq.username,
                                                                             password=State.config.bootstrap.rabbitmq.password)).init()
